/* Simple wavelet demo.
	 (c) Andreas Brentzen 2002,

	 This program is a very simple demo of wavelet encoding and decoding.
	 It does nothing fancy and is only intended as a demonstration of the
	 basic principles. It is not useful for anything practical.

	 The program should be called in the following way (no permutation of 
	 arguments :)

	 > wavelet COMMAND INFILE OUTFILE FRACTION

	 COMMAND  - either 'd' for decompose or 'r' - reconstruct
	 INFILE   - a PPM image in the case of 'd'. Otherwise a file
	            containing coefficients
	 OUTFILE  - coefficients in the case of 'd'. If we are reconstructing
	            it is the name of the reconstructed image.
   FRACTION - Use to specify the number of coefficients to retain, i.e.
	            .1 means "keep 10 percent".


*/

#include <cmath>
#include <vector>
#include <iostream>
#include <iomanip>

#include "PPMBitmap.h"
#include "Grid2D.h"

using namespace std;
using namespace GFX;
using namespace CMP;


/// Compute a mod b
inline int mod(int a, int b)
{
	a -= b * (a/b);
	if(a<0) a += b;
	return a;
}


/** Lookup is used to make a row or column of a 2D Grid look like
		a linear array. */
class LookUp
{
	float* data;
	int offset;
	int N;
	float dummy;
public:
	
	LookUp(float *_data, int _N, int _offset): 
		data(_data), N(_N), offset(_offset), dummy(.0f) {}

	void store(int i, float val)
	{
		if(i>=0 && i<N)
			data[i*offset] = val;
	}
	const float& operator[](int i) const
	{
		if(i<0||i>=N) return dummy;
		return data[i*offset];
	}
};



/** Time domain convolution filter class. Used for the CMF filters 
		of the fast wavelet transform. */
class Filter
{
	/// Number of coefficients ( N = E-B) 
	const int N;
	/// First coefficient
	const int B;
	/// Last coefficient
	const int E;
	std::vector<float> data;
	
public:
	
	int b() const {return B;}
	int e() const {return E;}
	int n() const {return N;}

	const float& operator[](int i) const 
	{
		return data[i-B];
	}
	
	Filter(int _N, int _B, int _E, const float* _data,float scaling=1.0f):
		N(_N), B(_B), E(_E), data(_data,_data+N) 
	{
		for(int i=0;i<data.size();++i)
			data[i] *= scaling;
	}

};


/// Create the bandpass filter G from the lowpass filter H.
inline const Filter create_bandpass(const Filter& h)
{
	int Bg = 1-h.e();
	int Eg = 1-h.b();
	vector<float> data(h.n());
	
	for(int i=Bg; i<=Eg; ++i)
		data[i-Bg]=pow(-1.0f,1-i)*h[1-i];

	return Filter(h.n(), Bg, Eg, &data[0]);
}


// A Bunch of filters ------------------------------


const float h_haar[2] = {1.0f/M_SQRT2, 1.0f/M_SQRT2};
const Filter Haar_H(2, 0, 1, h_haar);
const Filter Haar_G = create_bandpass(Haar_H);


const float h_7_5[7] = { -0.010714285714, 
												 -0.053571428571,  
												 0.260714285714,
												 0.607142857143,  
												 0.260714285714,
												 -0.053571428571,
												 -0.010714285714 };

const float h_7_5_d[5] = {-0.05, 0.25, 0.6, 0.25, -0.05};


const Filter S75_H(7, -3, 3, h_7_5,1);
const Filter S75d_H(5, -2, 2, h_7_5_d,2);
const Filter S75_G  = create_bandpass(S75d_H);
const Filter S75d_G = create_bandpass(S75_H);


// --------------------------------------------------

void decomposition_step(float* _data, int N, int offset)
{
	const Filter& H = S75d_H;
	const Filter& G = S75d_G;
//  	const Filter& H = Haar_H;
//  	const Filter& G = Haar_G;

	LookUp data(_data, N, offset);
	std::vector<float> a(N/2,0);
	std::vector<float> d(N/2,0);

	for(int i=0;i<N/2;++i)
		{
			int k = mod(2*i+H.b(),N);
			for(int j=H.b();j<=H.e();++j)
				{
					a[i] += data[k]*H[j];
					k = mod(k+1, N);
				}
			k = mod(2*i+G.b(),N);
			for(int j=G.b();j<=G.e();++j)
				{
					d[i] += data[k]*G[j];
					k = mod(k+1,N);
				}
		}
	for(int i=0;i<N/2;++i)
		{
			data.store(i, a[i]);
			data.store(N/2+i, d[i]);
		}

}

void reconstruction_step(float* _data, int N, int offset)
{
//   	const Filter& H = Haar_H;
//   	const Filter& G = Haar_G;
 	const Filter& H = S75_H;
 	const Filter& G = S75_G;

	LookUp data(_data, N, offset);

	std::vector<float> a(N,0);

	for(int i=0;i<N;++i)
		{
			int k = mod((i-H.b())/2,N/2);
			int m = H.b() + mod(i-H.b(),2);
			for(int j=m ;j<=H.e();j+=2)
				{
					a[i] += data[k]*H[j];
					k = mod(k-1, N/2);
				}
			k = mod((i-G.b())/2,N/2);
			m = G.b() + mod(i-G.b(),2);
			for(int j=m ;j<=G.e();j+=2)
				{
					a[i] += data[k+N/2]*G[j];
					k = mod(k-1, N/2);
				}
		}
	for(int i=0;i<N;++i)
		data.store(i,a[i]);

}

void decompose(float* data, int N)
{
	const float normalizer = sqrt(static_cast<float>(N));
	for(int i=0;i<N; ++i)
		data[i] /= normalizer;
	for(int g=N;g>=2;g/=2)
		decomposition_step(data,g,1);
}

void reconstruct(float* data, int N)
{
	for(int g=32; g<=N; g*= 2)
		reconstruction_step(data, g, 1);
	const float normalizer = sqrt(static_cast<float>(N));
	for(int i=0;i<N; ++i)
		data[i] *= normalizer;	
}


void decompose2D(Grid2D<float>& data, int N)
{
	const float normalizer = N;
	for(int i=0;i<N; ++i)
		for(int j=0;j<N; ++j)
			data(i,j) /= normalizer;
	for(int g=N;g>=2;g/=2)
		{
			for(int i=0;i<N;++i)
				decomposition_step(&data(0,i),g,1);
			for(int i=0;i<N;++i)
				decomposition_step(&data(i,0),g,N);
		}
}

void reconstruct2D(Grid2D<float>& data, int N)
{
	for(int g=2; g<=N; g*= 2)
		{
			for(int i=0;i<N;++i)
				reconstruction_step(&data(0,i),g,1);
			for(int i=0;i<N;++i)
				reconstruction_step(&data(i,0),g,N);
		}
	const float normalizer = N;
	for(int i=0;i<N; ++i)
		for(int j=0;j<N; ++j)
			data(i,j) *= normalizer;
}


bool check_dim(int& dim)
{
	if (dim < 2) 
		return false;
	static int powers2[12] = {2,4,8,16,32,64,128,256,512,1024,2048,4096};
	for(int i=0;i<12;++i)
		{
			if(powers2[i]==dim) return true;
			if(powers2[i]>dim) 
				{
					dim=powers2[i];
					return false;
				}
		}
	return false;
}

main(int argc, char **argv)
{
	string command = argv[1];
	string infile  = argv[2];
	string outfile = argv[3];
	
	if(command[0] == 'd')
		{
			float T;
			if(argc == 5) 
				{
					T = atof(argv[4]);
					cout << " threshold " << T << endl;
				}

			PPMBitmap picture_in(infile);
			int _XDIM = picture_in.get_width();
			int _YDIM = picture_in.get_height();
			int XDIM = _XDIM;
			int YDIM = _XDIM;

			check_dim(XDIM);
			check_dim(YDIM);

			const int DIM = max(XDIM,YDIM);
			const int sqdim = DIM*DIM;

			Grid2D<float> rchan(DIM,DIM,0);
			Grid2D<float> gchan(DIM,DIM,0);
			Grid2D<float> bchan(DIM,DIM,0);

			for(int i=0;i<_XDIM;++i)
				for(int j=0;j<_YDIM;++j)
					{
						rchan(i,j) = picture_in.get_pixel(i,j)[0];
						gchan(i,j) = picture_in.get_pixel(i,j)[1];
						bchan(i,j) = picture_in.get_pixel(i,j)[2];
					}

			decompose2D(rchan, DIM);
			decompose2D(gchan, DIM);
			decompose2D(bchan, DIM);

			std::vector<float> max_coeffs(DIM*DIM);
			for(int i=0;i<DIM;++i)
				for(int j=0;j<DIM;++j)
					max_coeffs[i*DIM+j] = max(fabs(rchan(i,j)),
																		max(fabs(gchan(i,j)),
																				fabs(bchan(i,j))));
			int n = (1.0f-T) * sqdim;
			nth_element(max_coeffs.begin(), max_coeffs.begin()+ n,
									max_coeffs.end());
			float thresh = max_coeffs[n];

 			int N=0;
			for(int i=0;i<DIM;++i)
				for(int j=0;j<DIM;++j)
					if(fabs(rchan(i,j)) < thresh &&
						 fabs(gchan(i,j)) < thresh &&
						 fabs(bchan(i,j)) < thresh)
						{
							rchan(i,j) = gchan(i,j) = bchan(i,j) = 0.0f;
							N++;
						}
					
			cout << "Coefficients retained = "
					 << (100*(sqdim - N)/sqdim) << " % " << endl;

			ofstream of(outfile.c_str(), ios::binary);
			of.write(reinterpret_cast<const char*>(&DIM), sizeof(int));
			of.write(reinterpret_cast<char*>(&rchan(0,0)), sizeof(float)*sqdim);
			of.write(reinterpret_cast<char*>(&gchan(0,0)), sizeof(float)*sqdim);
			of.write(reinterpret_cast<char*>(&bchan(0,0)), sizeof(float)*sqdim);
		}
	else
		{
			int DIM;
			ifstream ifs(infile.c_str(), ios::binary);
			ifs.read(reinterpret_cast<char*>(&DIM), sizeof(int));
			Grid2D<float> rchan(DIM,DIM,0);
			Grid2D<float> gchan(DIM,DIM,0);
			Grid2D<float> bchan(DIM,DIM,0);
			const int sqdim = DIM*DIM;

			ifs.read(reinterpret_cast<char*>(&rchan(0,0)), sizeof(float)*sqdim);
			ifs.read(reinterpret_cast<char*>(&gchan(0,0)), sizeof(float)*sqdim);
			ifs.read(reinterpret_cast<char*>(&bchan(0,0)), sizeof(float)*sqdim);
	
 			reconstruct2D(rchan, DIM);
 			reconstruct2D(gchan, DIM);
 			reconstruct2D(bchan, DIM);

			PPMBitmap reconst_picture(DIM, DIM);
			for(int i=0;i<DIM;++i)
				for(int j=0;j<DIM;++j)
					{
						ColorVector col(rchan(i,j), gchan(i,j),	bchan(i,j));
						reconst_picture.set_pixel(i,j,col);
					}
			
			reconst_picture.write(outfile);
		}

}
